/*
TBD/problems:

Categories to add:
- Bear # off
- others?


Questionable examples:
set gnubgid 4HPwATDgc/ABMA:MAEzAQAAAAAE
show z 24/18 13/9 
show z 24/14
doesn't distinguish. (Possible solution: print the unused segment.. a bit complicated)

set gnubgid yOfgAQPgc+IBAw:MAESANAASAAE 
show z 20/16(2) 13/9 (2)
"point" why not double point?? (One is advancing anchor - is this okay? Should it still be "double point"?)

set gnubgid mGfwATDgc/ABMA:cIkEAHAB+AAE 
show z 8/7(2) 6/5(2)
same problem



===============================================
Instructions: start up with "./gnubg -t", then  set up board "set gnubgid 4OvgASLgc/ABMA:8AlgACAAAAAE, then use the command "show z [move]"

Main external function:
MoveClassifyGenerateLabels(char *aLabel[], int ** aanMove, const TanBoard anBoard, const int size)


Main internal structures and functions and how they interact:
struct moveinfo
- This contains redundant but useful information about the move: the board after, whether it is contact.
- Also, whenever we test if the move is of a certain type, the result is stored here so we don't have to redo the computation.

struct classifiedchequermove
- Contains both the type of classification and up to 2 indices of segments that the move corresponds to.

moveinfo * CreateMoveInfo()
- Turns raw move data into a moveinfo object.

int IsThisMoveType(moveinfo *pmv, int iSegment[], const chequerclassification type)
- Based on the moveinfo given (pmv), determines whether the move is of the type specified by "type". If so, enters returns which segment(s) of the move in iSegment.
- It starts by checking if the computation has already been done (see moveinfo above); if not, it does the computation and stores the result in the moveinfo (as well as returning it).
- To compute whether the move is of the given type, there are a number of helper functions, such as Is...Sub() (most of these just determine whether a segment of a move indicates it being of a given type, and are lugged into IsGeneralContactType()).

classifiedchequermove * ClassifyChequerMove()
- Takes raw move data and returns an array of classifiedchequermove's
- It uses CreateMoveInfo to create the move
- Then goes through the possible classifications (in order) tests each one
- The returned array size is at most MAX_CLASSIFICATIONS.

void PrintMoveClassification()
- Prints a label about a given move classification, with a specified level of detail.
- Detail=0: just a single word or two describing the classification, e.g., "hit", "loose hit", "point", "run"
- Detail=1: also name the relevant pip, e.g., "hit 20", "loose hit 5", "7-point", "run from 24"
- Detail=2: name a second relevant pip if applicable, e.g. "hit 20 and 18", "loose hit 5 and 1", "7-,11-point".
   (Note many of these examples may not still hold in the future, when double hit, double point, double tiger, etc. are implemented.)

void PrintMultipleClassifications()
- Uses PrintMoveClassification() multiple times to print a label describing multiple classifications with "and" in between, e.g., "hit and run", "loose hit and split"
- Input includes an array of detail levels, where -1 in this array indicates not to print that part of the classification at all

void PrintMultipleClassificationArray()
- Uses PrintMoveClassification() multiple times along an array. See description there for more details.

void GenerateLabelsSub()
- Recursive function for generating unique labels.
- Input includes an integer array aPtn that partitions the given list, so if at one stage we get the same label for moves 0 and 2, then aPtn[0] and aPtn[2] will have the same value (see Repartition); this partition and value are fed into the next call of GenerateLabelsSub() so that these ones can be separated.
- See the comments at the code for Repartition() for more explanation of how that function works.

Here is an example of how it works.
Say gnubgid=4HPwASLgc/ABMA:cAklAQAAAAAE
 +13-14-15-16-17-18------19-20-21-22-23-24-+
 | X           O    |   | O              X |
 | X           O    |   | O              X |
 | X           O    |   | O                |
 | X                |   | O                |
 | X                |   | O                |
v|                  |BAR|                  |
 | O                |   | X                |
 | O                |   | X                |
 | O           X    |   | X                |
 | O           X    |   | X                | Rolled 21
 | O           X    |   | X     O        O |
 +12-11-10--9--8--7-------6--5--4--3--2--1-+

move[0] is 24/23 13/11, move[1] is 24/22 6/5*, move[2] is 24/23 6/4* (these are encoded through aanMove).

- ClassifyChequerMove will decide that move[0] is split, move[1] is loose hit (on 5), move[2] is loose hit (on 4)
- GenerateLabelsSub is called with aPtn = {0,0,0}, kVal=0, iteration=0
  - Comes up with labels "split", "loose hit", "loose hit" (level of detail based on DETAIL[0]={0,-1,-1})
  - Repartition will produce the partition {0,1,1} (indicating that the second and third are the same, and the first is different)
  - Since only one element of the partition has value 0, we are done with that label
  - Since >1 element has value 1, GenerateLabelsSub is called with aPtn = {0,1,1}, kVal=1, iteration=1
    - Comes up with labels "loose hit and split", "loose hit and split" (level of detail based on DETAIL[1]={0,0,-1})
    - Repartition produces the partition {-1,0,0} (indicating that the first is ignored, and the last two are the same)
    - Since >1 element has value 0, GenerateLabelsSub is called with aPtn = {-1,0,0}, kVal=0, iteration=2
      - Comes up with labels "loose hit 5", "loose hit 4" (level of detail based on DETAIL[2]={1,-1,-1})
      - Repartition produces the partition {-1,0,1} (indicating that the first is ignored, and the last two are different)
      - GenerateLabelsSub recognizes that each partitioned class has only 1 element so returns
- Now the labels are "split", "loose hit 5", "loose hit 4".
*/

#include <glib.h>
#include <glib.h>

#include "lib/gnubg-types.h"
#include "moveclassify.h"
#include "backgammon.h"
#include "drawboard.h"
#include "eval.h"

#define MAX_CLASSIFICATIONS (3) // Maximum number of classifications found using ClassifyChequerMove()

/* The different classification categories, in order of most to least relevant. The idea is that if a move 
satisfies more than one, then the highest ranked one is its "primary classification" and the next ranked one 
"secondary". (Hence this order only matters when comparing categories for the same move; many pairs of categories 
are mutually exclusive, so will never be compared.)
*/
typedef enum {
    BREAK_CONTACT=0,
    SHIFT, // not yet implemented
    LOOSE_DOUBLE_HIT, 
    MULTI_HIT,
    MULTI_POINT,
    DOUBLE_FALCON, 
    DOUBLE_TIGER,
    PICK_N_PASS,
    POINT_HIT,
    HIT_N_COVER, 
    SLOT, // Internal distinction with SAFE_SLOT
    POINT_MADE,
    LOOSE_HIT,
    HIT,
    CLEAR_LAST_POINT,  // Internal distinction with CLEAR_POINT
    ADVANCE_POINT,
    SPLIT,
    RUN,
    CLEAR_POINT, // Internal distinction with CLEAR_LAST_POINT
    TWO_UP,
    TWO_DOWN,
    BREAK_POINT,
    SAFE_SLOT, // Internal distinction with SLOT 
    NUM_CHEQUER_CLASSIFICATIONS
} chequerclassification;
 
typedef struct {
    chequerclassification classification;
    int iSegment[4];
    int numSegments;
} classifiedchequermove;

/* A move is determined by the initial board position (TanBoard before) and the chequer movements (anMove[8]).
The following contains that as well as a bunch of redundant data, which is useful for classifying the move.
*/
typedef struct {
    int anMove[8]; /* For each of up to 4 chequer movements, anMove[2*i] is initial pip (0 to 24) and anMove[2*i+1] is final pip (-1 to 23).
                      initial pip = -1 means end of movements. final pip = -1 means borne off.
                    */
    TanBoard before;   /* The board beforehand; this is int[2][25]; 
                          before[0][i] = number of opponent chequers on (i+1)th position (from their point of view), (before[0][24] = on bar).
                          before[1][i] = number of chequers on (i+1)th position.
                          Since each array is from that players point of view, before[0][i] corresponds to the same pip as before[1][23-i].
                          Chequers borne off are not directly tracked; to find them you must take 15-sum_i before[0][i].
                        */
    TanBoard after;    /* The board after applying the move. */
    int contactAfter;  /* Whether it is a contact position after the move. */
    
    int storedClass[NUM_CHEQUER_CLASSIFICATIONS]; /* Stores results of earlier computations. -1 means not yet computed. */
    classifiedchequermove storedClassData[NUM_CHEQUER_CLASSIFICATIONS];
} moveinfo;

typedef enum {BEFORE, AFTER, NONE} beforeorafter;

/* The possible points, ranked in order of value. (Somewhat ad hoc, and of course in reality the relative value of points depends highly on the position)
 Note all point values are shifted by -1; golden point=4, bar point=6, mid point=12, opp golden point= on bar=24. */
const int BEST_POINTS[] = {4, // Golden point
                         5, // 6
                         19,// Opp golden point
                         18,// Opp 6
                         3,20, // 4
                         6,17, // Bar points
                         7,16, // 8 points
                         2,21,1,22,0,23, // Other inner points
                         8,15,9,14,10,13,11,12, // Outer points, ending with mid
                         };

// Lookup table into BEST_POINTS array, i.e., BEST_POINTS[BEST_POINTS_LOOKUP[i]] = i.
const int BEST_POINT_LOOKUP[] = {14,12,10,4,0,1,6,8,16,18,20,22,23,21,19,17,9,7,3,2,5,11,13,15};

const char * POINT_NAMES[]={"off","1","2","3","4","5","6","7","8","9","10","11","12","13","14","15","16","17","18","19","20","21","22","23","24","bar"};

int IsThisMoveTypeSimple(moveinfo *pmv, const chequerclassification type);

typedef int (*determinersubfunc)(const moveinfo * pmv, int i);

int
IsGeneralContactType(const moveinfo *pmv, classifiedchequermove *pcl, const beforeorafter comparebasedon, determinersubfunc func) {
/* Tests if this move matches a category, where this can be tested by running func on all the sub-moves
*/
    if (!pmv->contactAfter)
        return FALSE;

    int pos[4]={30,30,30,30};
    pcl->numSegments=0;
    for (int i = 0; i < 8 && pmv->anMove[i] >= 0; i += 2)
        if ((*func)(pmv, i)) {
            int newPos=BEST_POINT_LOOKUP[comparebasedon==BEFORE ? pmv->anMove[i] : pmv->anMove[i+1]];
            int j;
            for (j=0;j<4 && newPos>pos[j];j++); // Find first j where newPos<=pos[j].
            if (j<4 && newPos!=pos[j]) { // Consequently, newPos<pos[j]
                pcl->numSegments++;
                for (int k=2; k>=j; k--) { // Shift later ones
                    pos[k+1]=pos[k];
                    pcl->iSegment[k+1]=pcl->iSegment[k];
                }
                pos[j]=newPos;
                pcl->iSegment[j]=i;
            } // Otherwise newPos==pos[j] or j==4, so do nothing
        }
    return pcl->numSegments>0;
}

int
FindEnd(const moveinfo *pmv, int i) {
/* Finds the pip where the move actually ends.
   E.g. for the move 24/18* /15, this would be broken into two segments; if i refers to the initial segment (24/18), 
   then this will return 15 (well actually 14, because of the shift by 1).
*/
    int iEnd=pmv->anMove[i+1];
    for (int j=i+2; j<8; j+=2)
        if (pmv->anMove[j]==iEnd)
            iEnd=pmv->anMove[j+1];
    return iEnd;
}

int
IsSlotSub(const moveinfo *pmv, int i){
// Determines whether the given partial move is a slot
#define MAX_SLOT_PIP 8
    if (pmv->anMove[i+1]< MAX_SLOT_PIP && pmv->anMove[i+1]>=0  // In the slotting range
                && pmv->after[1][pmv->anMove[i+1]]==1 // Leave a blot there
                && pmv->before[0][23-pmv->anMove[i+1]]==0) { // Not a loose hit
        if (pmv->anMove[i+1]>=3)
            return TRUE;
        else { // For points 3 or lower, only consider it a slot if the higher board points are made
            int areCovered=1;
            for (int j=pmv->anMove[i+1]+1; areCovered && j<=6; j++)
                areCovered=(pmv->after[1][j]>=2);
            return areCovered;
        }
    }
    return FALSE;
}

int
IsHitSub(const moveinfo *pmv, int i){
// Determines whether the given partial move is a hit
    return pmv->anMove[i+1]>=0 && pmv->anMove[i+1]<25 // Not on bar/bearing off
               && pmv->before[0][23-pmv->anMove[i+1]]==1; // Hit
}

int
IsLooseHitSub(const moveinfo *pmv, int i){
// Determines whether the given partial move is a loose hit
    return pmv->anMove[i+1]< 6 && pmv->anMove[i+1]>=0  // In the inner board
                && pmv->after[1][pmv->anMove[i+1]]==1 // Leave a blot there
                && IsHitSub(pmv, i);
}

int
IsAdvancePointSub(const moveinfo *pmv, int i){
// Determines whether the given partial move is advancing a point
    if (!(pmv->before[1][pmv->anMove[i]]-pmv->after[1][pmv->anMove[i]]>=2 // Move 2 chequers off the starting pip
             && pmv->after[1][pmv->anMove[i]] <2)) // The starting pip is no longer a point
        return FALSE;
    int iEnd=FindEnd(pmv,i);
    return iEnd != -1 // Didn't bear off
              && pmv->before[1][iEnd]==0 // Was not a point and ending pip before
              && pmv->after[1][iEnd]>=2; // Is a point there now
}

int
IsShiftSub(const moveinfo *pmv, int i){
// Determines whether the given partial move is a shift
    if (pmv->anMove[i]>=6  // Need to start in board
            || pmv->before[1][pmv->anMove[i]]<2 // Need a point at start
            || pmv->after[1][pmv->anMove[i]]>0 // Need start point to become empty
            || pmv->before[0][23-pmv->anMove[i+1]] != 1) // Need to hit a blot
        return FALSE;
    int iEnd=FindEnd(pmv,i);
    return pmv->before[1][iEnd]==0 && pmv->after[1][iEnd]>=2; // Check that the end went from empty to a point
}

int
IsPointMadeSub(const moveinfo *pmv, int i){
// Determines whether the given partial move is making a point
    return pmv->before[1][pmv->anMove[i+1]]<2  // No point there before
                && pmv->after[1][pmv->anMove[i+1]]>=2 // Leave a point there
                && !IsAdvancePointSub(pmv,i); // Not advancing a point
}

int
IsHitnCoverSub(const moveinfo *pmv, int i){
// Determines whether the given partial move is a hit and cover
    return pmv->anMove[i]<24 // Not starting on bar
                && pmv->before[0][23-pmv->anMove[i]]==1 // Hit (can't start move here, so necessarily this is a continuation) 
                && pmv->before[1][pmv->anMove[i+1]]==1 // Blot there before
                && pmv->after[1][pmv->anMove[i+1]]>=2; // Point there after
}

int
IsPointHitSub(const moveinfo *pmv, int i){
    return IsHitSub(pmv, i) && IsPointMadeSub(pmv, i);
}

int
IsRunSub(const moveinfo *pmv, int i){
// Determines whether the given partial move is a run
    if (pmv->anMove[i]<=17) // If not in the opp board/bar -> not run
        return FALSE;
    return !IsAdvancePointSub(pmv,i) && // Not advancing anchor
            FindEnd(pmv,i)<=17; // End up running -> is run
}

int
IsSplitSub(const moveinfo *pmv, int i){
// Determines whether the given partial move is splitting
#define MIN_SPLIT_PIP 16
    return pmv->anMove[i]>= MIN_SPLIT_PIP && pmv->anMove[i]<24  // In the split range
                && pmv->before[1][pmv->anMove[i]]==2 // Point there before
                && pmv->after[1][pmv->anMove[i]]==1 // Blot there after
                && pmv->after[1][pmv->anMove[i+1]]<2; // Not moving to safety
}

int
IsBreakPointSub(const moveinfo *pmv, int i){
// Determines whether the given partial move is splitting
    if (!(!IsSplitSub(pmv, i) // Not splitting 
                && pmv->before[1][pmv->anMove[i]]==2 // Point there before
                && pmv->after[1][pmv->anMove[i]]==1)) // Blot there after
        return FALSE;
    for (int j=23-pmv->anMove[i]; j<25; j++)
        if (pmv->after[0][j])
            return TRUE;    // Check that the opponent has a chequer beyond the point being broken
    return FALSE;
}

int
IsUpSub(const moveinfo *pmv, int i) {
    return pmv->anMove[i]>= 18 && pmv->anMove[i]<24;  // In the opp board
}

int
IsDownSub(const moveinfo *pmv, int i){
// Determine whether the partial move is "bringing one down"
   return pmv->anMove[i]==12 || pmv->anMove[i]==13;  // In the bringing down range
}

int
IsContact (const TanBoard anBoard) {
/* Determines whether this is a contact position. */
    int lastChequer;
    int opp;
   
    for (lastChequer = 24; lastChequer >= 0 && anBoard[0][lastChequer]==0; --lastChequer); // Find last chequer
    for (opp=24-lastChequer; opp<25 && anBoard[1][opp]==0; opp++); // Look for an opponent chequer beyond this position
    return (opp<25);
}

int
IsBreakContact (const moveinfo * pmv, classifiedchequermove * pcl) {
/* Determines whether this move is considered breaking contact
*/
    pcl->numSegments=0;
    return !pmv->contactAfter && IsContact(pmv->before);
}

int
IsPicknPassSub(const moveinfo * pmv, int i) {
// Determines whether the given partial move is a pick n pass

    if (!IsHitSub(pmv, i)) // Pick (hit)
        return FALSE;
    return pmv->anMove[i]<=5 && // Hit in inner board
        (pmv->before[1][FindEnd(pmv,i)]>=2); // Pass (end on an existing point)
}

int
IsClearPointSub(const moveinfo *pmv, int i) {
// Determines whether the given partial move is clearing a point
    return (!IsAdvancePointSub(pmv, i) // Not advancing the point
            && pmv->anMove[i]<24 // Not the bar
            && pmv->before[1][pmv->anMove[i]]>=2 // A point before
            && pmv->after[1][pmv->anMove[i]]==0); // Empty after
}

int
IsDoubleFalconBoard(const TanBoard after) {
// Checks whether the "after" board is consistent with a "double falcon" move
    if (after[0][24]>0) return FALSE; // Opp on bar
    int blot1, blot2;
    for (blot1=15; blot1<18 && after[1][blot1]!=1; blot1++);
    for (blot2=blot1+1; blot2<18 && after[1][blot2]!=1; blot2++);
    if (blot2==18) return FALSE; // Only one blot
    int count=0; // Count number of opp pips that can hit both
    for (int i=23-blot1-1; i<=23-blot2+6; i++)
        if (after[0][i]>0) count++;
    if (count>1) return FALSE; // Too many
    int oppPt;
    for (oppPt=23-blot1-1; oppPt<=23-blot1+6 && after[0][oppPt]<2; oppPt++); // Find an opp point in the 12-18 point range
    if (oppPt=23-blot1+6 // No opp point in the relevant range
            || after[0][oppPt]>2) // Not stripped
        return FALSE; 
    return TRUE;
}

int
IsThisMoveType(moveinfo *pmv, classifiedchequermove *pcl) {
    if (pmv->storedClass[pcl->classification]!=-1) {
        if (pmv->storedClass[pcl->classification])
            memcpy(pcl, &pmv->storedClassData[pcl->classification],sizeof(classifiedchequermove));
        return pmv->storedClass[pcl->classification];
    }
    pcl->numSegments=0;
    int result=FALSE;
    switch (pcl->classification) {
    case CLEAR_POINT:
    case CLEAR_LAST_POINT:
        if (IsThisMoveTypeSimple(pmv, TWO_UP) || !IsGeneralContactType(pmv, pcl, BEFORE, *IsClearPointSub))
            break;
        int maxPtCleared=pmv->anMove[pcl->iSegment[0]];
        for (int i=1; i<pcl->numSegments; i++) {
            if (pmv->anMove[pcl->iSegment[i]] > maxPtCleared && pcl->classification==CLEAR_LAST_POINT) {
                maxPtCleared=pcl->iSegment[0]; // Ensure that iSegment[0] is for the biggest point cleared
                pcl->iSegment[0]=pcl->iSegment[i];
                pcl->iSegment[i]=maxPtCleared;
            } else
                maxPtCleared=MAX(maxPtCleared, pcl->iSegment[i]);
        }
        int i;
        for (i=maxPtCleared+1; i<25 && pmv->after[1][i]==0; i++); // Find if there are any chequers beyond this point
        result= (i==25)==(pcl->classification==CLEAR_LAST_POINT);
        break;
    case DOUBLE_FALCON:
        if (!IsDoubleFalconBoard(pmv->after) || !IsGeneralContactType(pmv,pcl, AFTER, *IsRunSub) || pcl->numSegments<2)
            break;
        result=(pmv->anMove[pcl->iSegment[0]+1]>=15 && pmv->anMove[pcl->iSegment[1]+1]>=15);
        break;
    case SLOT:
    case SAFE_SLOT:
        if (!IsGeneralContactType(pmv, pcl, AFTER, *IsSlotSub)) 
            break;
        int isSafe=TRUE;
        for (int i=0; i<pmv->anMove[pcl->iSegment[0]+1] && isSafe; i++) 
             isSafe=(pmv->after[0][23-i]==0);
        result=(isSafe==(pcl->classification==SAFE_SLOT));
        break;
    case LOOSE_DOUBLE_HIT: {
        classifiedchequermove cl2;
        result=!IsThisMoveTypeSimple(pmv,DOUBLE_TIGER) &&
               IsGeneralContactType(pmv, pcl, AFTER, *IsHitSub) && pcl->numSegments>=2 &&
               IsGeneralContactType(pmv, &cl2, AFTER, *IsLooseHitSub);
        if (result) {
            for (int i=0; i<pcl->numSegments; i++) // Put loose hit first
                if (pmv->anMove[pcl->iSegment[i]+1]<=5) {
                    int tmp=pcl->iSegment[0];
                    pcl->iSegment[0]=pcl->iSegment[i];
                    pcl->iSegment[i]=tmp;
                    break;
                }
            }
        }
        break;
    case LOOSE_HIT:
        result=!IsThisMoveTypeSimple(pmv,DOUBLE_TIGER) && !IsThisMoveTypeSimple(pmv,LOOSE_DOUBLE_HIT) &&
               IsGeneralContactType(pmv, pcl, AFTER, *IsLooseHitSub);
        break;
    case DOUBLE_TIGER:
        result=IsGeneralContactType(pmv,pcl, AFTER, *IsLooseHitSub) && pcl->numSegments>=2;
        break;
    case HIT:
        result=!IsThisMoveTypeSimple(pmv,LOOSE_HIT) && !IsThisMoveTypeSimple(pmv,POINT_HIT) && 
               !IsThisMoveTypeSimple(pmv,MULTI_HIT) && !IsThisMoveTypeSimple(pmv,DOUBLE_TIGER) && 
               !IsThisMoveTypeSimple(pmv,LOOSE_DOUBLE_HIT) &&
               IsGeneralContactType(pmv, pcl, AFTER, *IsHitSub);
        break;
    case MULTI_HIT:
        result=!IsThisMoveTypeSimple(pmv,DOUBLE_TIGER) &&
               IsGeneralContactType(pmv,pcl, AFTER, *IsHitSub) && pcl->numSegments>=2;
        if (result && IsThisMoveTypeSimple(pmv,LOOSE_DOUBLE_HIT))
            result=(pcl->numSegments>=3); // Check for triple hit
        break;
    case HIT_N_COVER:
        result=IsGeneralContactType(pmv, pcl, AFTER, *IsHitnCoverSub);
        break;
    case POINT_MADE:
        result=!IsThisMoveTypeSimple(pmv,ADVANCE_POINT) && !IsThisMoveTypeSimple(pmv,POINT_HIT) 
               && !IsThisMoveTypeSimple(pmv,MULTI_POINT) &&
               IsGeneralContactType(pmv, pcl, AFTER, *IsPointMadeSub);
        break;
    case MULTI_POINT:
        result=IsGeneralContactType(pmv,pcl, AFTER, *IsPointMadeSub) && pcl->numSegments>=2;
        break;
    case SPLIT:
        result=!IsThisMoveTypeSimple(pmv,RUN) && !IsThisMoveTypeSimple(pmv,TWO_UP) &&
               IsGeneralContactType(pmv, pcl, BEFORE, *IsSplitSub);
        break;
    case RUN:
        result=IsGeneralContactType(pmv, pcl, BEFORE, *IsRunSub);
        pcl->numSegments=MIN(1,pcl->numSegments); // Ignore second one; it may be a remnant of the first
        break;
    case TWO_DOWN:
        result=IsGeneralContactType(pmv,pcl, AFTER, *IsDownSub) && pcl->numSegments>=2;
        break;
    case TWO_UP:
        result=IsGeneralContactType(pmv,pcl, AFTER, *IsUpSub) && pcl->numSegments>=2;
        break;
    case BREAK_CONTACT:
        result=IsBreakContact(pmv, pcl);
        break;
    case PICK_N_PASS:
        result=IsGeneralContactType(pmv, pcl, AFTER, *IsPicknPassSub);
        break;
    case ADVANCE_POINT:
        result=!IsThisMoveTypeSimple(pmv,SHIFT) &&
                IsGeneralContactType(pmv, pcl, AFTER, *IsAdvancePointSub);
        break;
    case POINT_HIT:
        result=IsGeneralContactType(pmv, pcl, AFTER, *IsPointHitSub);
        break;
    case BREAK_POINT:
        result=IsGeneralContactType(pmv, pcl, BEFORE, *IsBreakPointSub);
        break;
    case SHIFT:
        result=IsGeneralContactType(pmv, pcl, BEFORE, *IsShiftSub);
        break;
    default:
        result=FALSE;
    }
    if (pcl->classification>=0 && pcl->classification<NUM_CHEQUER_CLASSIFICATIONS) {
        pmv->storedClass[pcl->classification]= result ? 1 : 0;
        if (result) {
            memcpy(&pmv->storedClassData[pcl->classification], pcl, sizeof(classifiedchequermove));
        }
    }
    return result;
}

int
IsThisMoveTypeSimple(moveinfo *pmv, const chequerclassification type) {
    classifiedchequermove cl;
    cl.classification=type;
    return IsThisMoveType(pmv, &cl);
}


moveinfo *
CreateMoveInfo(const int anMove[8], const TanBoard anBoard) {
/* Creates a moveinfo from the given board and move combo.
WARNING: This allocates memory. Must be freed elsewhere.
*/
    moveinfo *pmv = g_malloc(sizeof(moveinfo));
    memcpy(pmv->before,anBoard,sizeof(pmv->before));
    memcpy(pmv->anMove,anMove,sizeof(pmv->anMove));
    memcpy(pmv->after,anBoard,sizeof(pmv->after));
    if (ApplyMove(pmv->after, anMove, FALSE)) {
        g_free(pmv);
        return NULL;
    }
    pmv->contactAfter=IsContact(pmv->after);
    for (int i=0; i<NUM_CHEQUER_CLASSIFICATIONS; i++)
        pmv->storedClass[i]=-1;

    return pmv;
}

classifiedchequermove *
ClassifyChequerMove(const int anMove[8], const TanBoard anBoard, int * numClassifications) {
/* Classifies the given move. If the move does not make sense, returns NULL. Otherwise returns an array of classifications, of size numClassifications.
WARNING: This allocates memory. Must be freed elsewhere.
*/
    moveinfo *pmv = CreateMoveInfo(anMove, anBoard);
    if (pmv==NULL)
        return NULL;
    classifiedchequermove *result=g_malloc(sizeof(classifiedchequermove) * MAX_CLASSIFICATIONS);
    *numClassifications=0;
    for (int i=0; i<NUM_CHEQUER_CLASSIFICATIONS && *numClassifications<MAX_CLASSIFICATIONS; i++) {
        result[*numClassifications].classification=i;
        if ( IsThisMoveType(pmv, &result[*numClassifications]) )
            (*numClassifications)++;
    }
    g_free(pmv);
    return result;
}

#define MAX_STRING_LENGTH 30 // Actually only 23
static const char * MULTI_NAME_STR[5]={"","single"/*not used*/,"double","triple","quadruple"};


void
PrintMoveClassification(char *sz, const classifiedchequermove *pcl, const int anMove[8], int detail) {
/* detail=0, 1, or 2 (shortest to longest)
*/
    int before1=0, before2=0, after1=0, after2=0;
    detail=MIN(detail,pcl->numSegments);
    if (pcl->numSegments>=1) {
        before1=anMove[pcl->iSegment[0]]+1;
        after1=anMove[pcl->iSegment[0]+1]+1;
    }
    if (pcl->numSegments>=2) {
        before2=anMove[pcl->iSegment[1]]+1;
        after2=anMove[pcl->iSegment[1]+1]+1;
    }
    switch (pcl->classification) {
    case SLOT:
    case SAFE_SLOT:
        switch (detail) {
        case 2:
            sprintf(sz,"slot %d,%d", after1, after2);
            break;
        case 1:
            sprintf(sz,"slot %d", after1);
            break;
        default:
            sprintf(sz,"slot");
        }
        break;
    case LOOSE_HIT:
        switch (detail) {
        case 2:
            sprintf(sz,"loose hit %d,%d", after1, after2);  // Shouldn't occur
            break;
        case 1:
            sprintf(sz,"loose hit %d", after1);
            break;
        default:
            sprintf(sz,"loose hit");
        }
        break;
    case HIT:
        switch (detail) {
        case 2:
            sprintf(sz,"hit %d,%d", after1, after2); // Shouldn't occur 
            break;
        case 1:
            sprintf(sz,"hit %d", after1);
            break;
        default:
            sprintf(sz,"hit");
        }
        break;
    case POINT_MADE:
        switch (detail) {
        case 2:
            sprintf(sz,"%d- and %d-point", after1, after2); // Shouldn't occur
            break;
        case 1:
            if (after1>=17)
                sprintf(sz,"%d-anchor", after1);
            else
                sprintf(sz,"%d-point", after1);
            break;
        default:
            if (after1>=17)
                sprintf(sz,"anchor");
            else
                sprintf(sz,"point");
        }
        break;
    case RUN:
        switch (detail) {
        case 2:
            sprintf(sz,"run from %s,%s", POINT_NAMES[before1], POINT_NAMES[before2]);
            break;
        case 1:
            sprintf(sz,"run from %s", POINT_NAMES[before1]);
            break;
        default:
            sprintf(sz,"run");
        }
        break;
    case SPLIT:
        switch (detail) {
        case 2:
            sprintf(sz,"split %d and %d", before1, before2); // Shouldn't occur (?)
            break;
        case 1:
            sprintf(sz,"split from %d", before1);
            break;
        default:
            sprintf(sz,"split");
        }
        break;
    case TWO_DOWN:
        sprintf(sz, "two down");
        break;
    case TWO_UP:
        sprintf(sz, "two up");
        break;
    case ADVANCE_POINT:
        switch (detail) {
        case 2:
            sprintf(sz,"advance %d-,%d-point", before1, before2); 
            break;
        case 1:
            if (before1 >= 17)
                sprintf(sz,"advance anchor to %d", after1);
            else
                sprintf(sz,"advance %d-point", before1);
            break;
        default:
            if (before1 >= 17 && pcl->numSegments==1)
                sprintf(sz,"advance anchor");
            else if (before1 >= 17 && before2>=17 && pcl->numSegments==2)
                sprintf(sz,"advance anchors");
            else if (pcl->numSegments>=2)
                sprintf(sz,"advance points");
            else
                sprintf(sz,"advance point");
        }
        break;
    case POINT_HIT:
        switch (detail) {
        case 2:
            sprintf(sz,"point-hit on %d,%d", after1, after2);
            break;
        case 1:
            if (before1 >= 17)
                sprintf(sz,"anchor-hit on %d", after1);
            else
                sprintf(sz,"point-hit on %d", after1);
            break;
        default:
            if (before1 >= 17)
                sprintf(sz,"anchor on head");
            else if (pcl->numSegments>=2)
                sprintf(sz,"points on head");
            else
                sprintf(sz,"point on head");
        }
        break;
    case MULTI_HIT:
    case MULTI_POINT:
    case DOUBLE_TIGER: {
        char *szType = (pcl->classification==MULTI_HIT) ? "hit" 
                       : ( (pcl->classification==DOUBLE_TIGER) ? "tiger" 
                       : ( (pcl->numSegments==2 && MIN(after1,after2)>=17) ? "anchor" 
                       : "point" ) );
        switch (detail) {
        case 1: // Give more detail than requested... but what else can we do?
        case 2:
            switch (pcl->numSegments) {
            case 4:
                sprintf(sz,"%s %s (%d,%d,%d,%d)", MULTI_NAME_STR[pcl->numSegments], szType, after1, after2, anMove[pcl->iSegment[2]+1], anMove[pcl->iSegment[3]+1]);
                break;
            case 3:
                sprintf(sz,"%s %s (%d,%d,%d)", MULTI_NAME_STR[pcl->numSegments], szType, after1, after2, anMove[pcl->iSegment[2]+1]);
                break;
            default: // 2
                sprintf(sz,"%s %s (%d,%d)", MULTI_NAME_STR[pcl->numSegments], szType, after1, after2);
            }
            break;
        default:
            sprintf(sz,"%s %s", MULTI_NAME_STR[pcl->numSegments],szType);
        }
        }
        break;
    case DOUBLE_FALCON:
        sprintf(sz,"double falcon");
        break;
    case BREAK_CONTACT:
        sprintf(sz,"break contact");
        break;
    case HIT_N_COVER:
        switch (detail) {
        case 1: 
        case 2:
            sprintf(sz,"hit %d and cover %d", before1, after1); 
            break;
        default:
            sprintf(sz,"hit and cover");
        }
        break;
    case BREAK_POINT:
        switch (detail) {
        case 2:
            sprintf(sz,"break %d,%d-points", before1, before2);
            break;
        case 1:
            if (before1 >= 16)
                sprintf(sz,"break anchor (%d)", before1);
            else
                sprintf(sz,"break %d-point", before1);
            break;
        default:
            if (before1 >= 16 && pcl->numSegments==1)
                sprintf(sz,"break anchor");
            else if (before1 >= 16 && pcl->numSegments==2 && before2>=16)
                sprintf(sz,"break anchors");
            else if (pcl->numSegments>=2)
                sprintf(sz,"break points");
            else
                sprintf(sz,"break point");
        }
        break;
    case PICK_N_PASS:
        switch (detail) {
        case 2:
        case 1:
            sprintf(sz,"pick n pass (hit %d)", after1);
            break;
        default:
            sprintf(sz,"pick n pass");
        }
        break;
    case LOOSE_DOUBLE_HIT:
        switch (detail) {
        case 1: // Give more detail than requested... but what else can we do?
        case 2:
            sprintf(sz,"hit %d and loose hit %d", after2, after1); 
            break;
        default:
            sprintf(sz,"loose %s hit", MULTI_NAME_STR[pcl->numSegments]);
        }
        break;
    case SHIFT:
        switch (detail) {
        case 1: 
        case 2:
            sprintf(sz,"shift (hit %d)", after1);
            break;
        default: 
            sprintf(sz,"shift");
        }
        break;
    case CLEAR_LAST_POINT:
    case CLEAR_POINT:
        switch (detail) {
        case 2:
            sprintf(sz,"clear %d,%d-points", before1, before2);
            break;
        case 1:
            sprintf(sz,"clear %d-point", before1);
            break;
        default:
            sprintf(sz,"clear point");
        }
        break;
    default:
        sprintf(sz,"??"); // Should not occur.
    }
}


void
Repartition(/*const*/ char * aLabel[], const int aPtnIn[], const int kVal, const int size, int aPtnOut[]) {
/* Repartitions the elements of aLabel which are already in the partition segment with value kVal.
For i in {0,..,size-1} for which aPtnIn[i]==kVal, we assign positive values to aPtnOut[i] such that
aPtnOut[i]==aPtnOut[j] iff aLabel[i], aLabel[j] represent the same string.
If aPtn[i]!=kVal, aPtnOut[i] is assigned -1.
Input:
aLabel[size]: An array of strings.
aPtnIn[size]: An existing partition.
kVal: The value that is selected using aPtnIn.
size: The size of the arrrays
aPtnOut[size]: An array that is written to, defining a subpartition.
*/
    int i=0;
    for (i=0; i<size; i++)
        aPtnOut[i]=-1; // Initialize: set all relevant values to -1
    int newVal=0;
    i=0;
    while (TRUE) {
        for (; i<size && !(aPtnIn[i]==kVal && aPtnOut[i]==-1); i++); // Find next entry that hasn't been dealt with yet
        if (i==size) // Done
            break;
        for (int j=i; j<size; j++)
            if (strcmp(aLabel[i], aLabel[j])==0) // Same string
                aPtnOut[j]=newVal;
        newVal++;
    }

}

void
PrintMultipleClassifications(char *sz, const int anMove[8], const classifiedchequermove aCl[], const int aClSize, const int detail[MAX_CLASSIFICATIONS]) {
/* Prints move classification to sz for the given array of classifications. detail[l] specifies the amount of detail to print for the l^th classification:
If detail[l]==-1: don't print this level at all; otherwise, prints this classification with detail[i] amount of detail.
It is assumed that detail[] is non-increasing (in particular, if detail[l]==-1, this does not go on to higher values of l).
Input:
anMove[8]: The move.
aCl[aClSize]: Classifications.
aClSize: Number of classifications. Must be >0 and must not exceed MAX_CLASSIFICATIONS.
detail[NUM_CHEQUER_CLASSIFICATIONS]: Level of detail to print at each level.
Output:
sz: char * to print the strings. Should have size at least MOVE_CLASSIFICATION_LABEL_SIZE.
*/
    PrintMoveClassification(sz, &aCl[0], anMove, detail[0]);
    char buf[MAX_STRING_LENGTH];
    for (int j=1; j<aClSize && detail[j]>-1; j++) {
        strcat(sz, " and ");
        PrintMoveClassification(buf, &aCl[j], anMove, detail[j]);
        strcat(sz, buf);
    }
}

void
PrintMultipleClassificationsArray(char *aLabel[], /*const*/ int ** aanMove, /*const*/ classifiedchequermove * aaCl[], const int aClSize[], const int aPtn[], const int kVal, const int detail[MAX_CLASSIFICATIONS], const int size) {
/* Prints move classification to aLabel[i], for all i for which aPtn[i]==kVal. detail[l] specifies the amount of detail to print for the l^th classification:
If detail[l]==-1: don't print this level at all; otherwise, prints this classification with detail[i] amount of detail.
It is assumed that detail[] is non-increasing (in particular, if detail[l]==-1, this does not go on to higher values of l).
Input:
aanMove[size][8]: Moves.
aaCl[size][]: Classifications.
aClSize[size]: Number of classifications; that is, apCl[i] is of size apCl[i], for each i. If apCl[i]==0 then no change is made to aLabel[i]. apCl[i] should not exceed MAX_CLASSIFICATIONS.
detail[NUM_CHEQUER_CLASSIFICATIONS]: Level of detail to print at each level.
aPtn[size]: A partition of the indices; used to select which ones to print.
kVal: The value to select from the partition.
size: Array size.
Output:
aLabel[size]: char * to print the strings. Each should have size at least MOVE_CLASSIFICATION_LABEL_SIZE.
*/
    if (detail[0]==-1)
        return;
    for (int i=0; i<size; i++)
        if (aPtn[i]==kVal && aClSize[i]>0)
            PrintMultipleClassifications(aLabel[i], aanMove[i], aaCl[i], aClSize[i], detail);
}

#define MAX_ITERATE 9
const int DETAIL[MAX_ITERATE][MAX_CLASSIFICATIONS] = {
{0,-1,-1},
{0,0,-1},
{1,-1,-1},
{1,0,-1},
{1,1,-1},
{2,-1,-1},
{2,0,-1},
{2,1,-1},
{2,2,-1}
};

void
GenerateLabelsSub(char *aLabel[], /*const*/ int ** aanMove, const TanBoard anBoard, /*const*/ classifiedchequermove * apCl[], const int aClSize[], const int aPtn[], const int kVal, const int size, const int iteration) {
    if (iteration>MAX_ITERATE) {
        // Last iteration; can't improve, so default with FormatMove
        for (int i=0; i<size; i++)
            if (aPtn[i]==kVal) {
                FormatMove(aLabel[i], anBoard, aanMove[i]);
            }
        return;
    }
    // Attempt to print levels according to current level of detail
    PrintMultipleClassificationsArray(aLabel, aanMove, apCl, aClSize, aPtn, kVal, DETAIL[iteration], size);
    int aPtn2[size];
    Repartition(aLabel, aPtn, kVal, size, aPtn2);
    for (int k=0; k<size; k++) {
        // Determine how many are in the k-th partition
        int curNum=0;
        for (int i=0; i<size; i++) { 
            if (aPtn2[i]==k)
                curNum++;
        }
        if (curNum>1) // Current detail level did not distinguish these ones; move on to the next level
            GenerateLabelsSub(aLabel, aanMove, anBoard, apCl, aClSize, aPtn2, k, size, iteration+1);
    }
}



void
MoveClassifyGenerateLabels(char *aLabel[], /*const*/ int ** aanMove, const TanBoard anBoard, const int size) {
/* Generates labels for the given moves that are:
- Distinct (definitely)
- Descriptive (as much as possible)
- Concise (as much as possible)

Input:
aLabel[size]: char * in which to print the labels
anMove[num][8] : aanMove[i] is a move, for each i.
anBoard : The board (before applying move).
num : The number of moves. Note: this function may work for any value, but is designed with num around 3 or less.
Output:
aLabel[size]: char * to print the strings. Each should have size at least MOVE_CLASSIFICATION_LABEL_SIZE.
*/
    classifiedchequermove * apCl[size];
    int aClSize[size];
    int aPtn[size];
    for (int i=0; i<size; i++) {
        apCl[i]=ClassifyChequerMove(aanMove[i], anBoard, &aClSize[i]);
        if (apCl[i]==NULL) { // Problem with the move!
            strcpy(aLabel[i],"");
            aPtn[i]=-1;
        } else if (aClSize[i]==0) { // No classification - just use normal formatting
            FormatMove(aLabel[i], anBoard, aanMove[i]);
            aPtn[i]=-1;
        } else
            aPtn[i]=0;
    }
    GenerateLabelsSub(aLabel, aanMove, anBoard, apCl, aClSize, aPtn, 0, size, 0);
    for (int i=0; i<size; i++)
        g_free(apCl[i]);
}

void
MoveClassifyGenerateLabel(char *sz, const int anMove[8], const TanBoard anBoard) {
    int size;
    classifiedchequermove * pCl=ClassifyChequerMove(anMove, anBoard, &size);
    if (pCl != NULL && size>0)
        PrintMoveClassification(sz, pCl, anMove, 1);
    else
        strcpy(sz,"");
    if (pCl!=NULL)
        g_free(pCl);
}


// static variables used to test.
int * pastThreeMoves[3] = {NULL, NULL, NULL};
int pastThreeMovesSize=0;


void
TestClassMove(const int anMove[8], const TanBoard anBoard) {

    int numClassifications;
    classifiedchequermove *cl = ClassifyChequerMove(anMove, anBoard, &numClassifications);
    if (cl==NULL) {
        g_print("Illegal move or no board\n");
        return;
    }
    if (numClassifications==0)
        g_print("No classification\n");
    for (int i=0; i<numClassifications; i++)
    {
        char sz[100];
        PrintMoveClassification(sz,&cl[i],anMove,2);
        g_print("%s\n",sz);
    }
    g_free(cl);

    if (pastThreeMovesSize==3)
        for (int i=0; i<2; i++)
            memcpy(pastThreeMoves[i],pastThreeMoves[i+1], 8*sizeof(int));
    else {
        pastThreeMoves[pastThreeMovesSize]=g_malloc(8*sizeof(int));
        pastThreeMovesSize++;
    }
    memcpy(pastThreeMoves[pastThreeMovesSize-1], anMove, sizeof(int)*8);
    char * aLabel[pastThreeMovesSize];
    for (int i=0; i<pastThreeMovesSize; i++)
        aLabel[i]=g_malloc(MOVE_CLASSIFICATION_LABEL_SIZE*sizeof(char));
    MoveClassifyGenerateLabels(aLabel, pastThreeMoves, anBoard, pastThreeMovesSize);
    for (int i=0; i<pastThreeMovesSize; i++) {
        char sz[100];
        FormatMove(sz, anBoard, pastThreeMoves[i]);
        g_print("Move %s: %s\n",sz, aLabel[i]);
        g_free(aLabel[i]);
    }
}

// Copied from play.c
static void
parsemove_to_anmove(int c, int anMove[])
{
    int i;
    for (i = 0; i < 4; i++) {
        if (i < c) {
            anMove[2*i]--;
            anMove[2*i + 1]--;
        } else {
            anMove[2*i] = -1;
            anMove[2*i + 1] = -1;
        }
    }
    CanonicalMoveOrder(anMove);
}

extern void
CommandShowClassifyTest(char *sz) {
/*    classifiedchequermove cl;
    int an[8]={-1,-1,-1,-1,-1,-1,-1,-1};
    char ssz[MOVE_CLASSIFICATION_LABEL_SIZE];
    for (cl.classification=0; cl.classification<NUM_CHEQUER_CLASSIFICATIONS; cl.classification++) {
        PrintMoveClassification(ssz,&cl,an,0);
        g_print("%s\n",ssz);
    }
*/      

    int an[8];
    int c=ParseMove(sz, an);
    if (c<0) {
        g_print("Could not parse move");
        return;
    }
//    g_print("c=%d ",c);
//    for (int i=0; i<8; i++) g_print("%d ",an[i]);
    parsemove_to_anmove(c,an);
    g_print("\n");
    TestClassMove(an, msBoard());
}

